#include <windows.h>
#include <math.h>
#include <string.h>
#include "usersurf.h"

/*
Written by Kenneth E. Moore
April 22, 1997
Modified August 2, 2002 to support scaling factors

This DLL models an arbitrary number of cylindrical lens elements in a vertical lens array.

The user provides the number of elements in y, the size in y of
each element, and the radius, conic, and glass.

This surface breaks up the beam into numerous separate
beams, and so most ZEMAX features will fail to work with this surface.

However, the spot diagrams, image analysis, etc, all work okay.



*/


int __declspec(dllexport) APIENTRY UserDefinedSurface(USER_DATA *UD, FIXED_DATA *FD);
int GetCellCenter(int ny, double wy, double y, double *cy);

/* a generic Snells law refraction routine */
int Refract(double thisn, double nextn, double *l, double *m, double *n, double ln, double mn, double nn);

BOOL WINAPI DllMain (HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
	{
   return TRUE;
   }

/* this DLL models a lens array surface type */

int __declspec(dllexport) APIENTRY UserDefinedSurface(USER_DATA *UD, FIXED_DATA *FD)
	{
   int i, ny, error, miss_flag;
   double p2, alpha, xpower, ypower, a, b, c, rad, casp, t, zc;
   double wy, cy, x, y, z;
   switch(FD->type)
   	{
      case 0:
      	/* ZEMAX is requesting general information about the surface */
         switch(FD->numb)
         	{
            case 0:
            	/* ZEMAX wants to know the name of the surface */
		         /* do not exceed 12 characters */
		         strcpy(UD->string,"Cyl Array");
               break;
            case 1:
            	/* ZEMAX wants to know if this surface is rotationally symmetric */
               /* it is not, so return a null string */
            	UD->string[0] = '\0';
               break;
            case 2:
            	/* ZEMAX wants to know if this surface is a gradient index media */
               /* it is not, so return a null string */
            	UD->string[0] = '\0';
            	break;
            }
         break;
      case 1:
      	/* ZEMAX is requesting the names of the parameter columns */
         /* the value FD->numb will indicate which value ZEMAX wants. */
         /* returning a null string indicates that the parameter is unused. */
         switch(FD->numb)
         	{
            case 1:
            	strcpy(UD->string, "Number Y");
               break;
            case 2:
            	strcpy(UD->string, "Height Y");
               break;
            default:
            	UD->string[0] = '\0';
            	break;
            }
      	break;
      case 2:
      	/* ZEMAX is requesting the names of the extra data columns */
         /* the value FD->numb will indicate which value ZEMAX wants. */
         /* they are all "Unused" for this surface type */
         /* returning a null string indicates that the extradata value is unused. */
         switch(FD->numb)
         	{
            default:
            	UD->string[0] = '\0';
            	break;
            }
      	break;
      case 3:
      	/* ZEMAX wants to know the sag of the surface */
         /* if there is an alternate sag, return it as well */
         /* otherwise, set the alternate sag identical to the sag */
         /* The sag is sag1, alternate is sag2. */

         UD->sag1 = 0.0;
         UD->sag2 = 0.0;

			/* if a plane, just return */
			if (FD->cv == 0) return(0);

         /* figure out the center coordinates of which "cell" we are in */
         ny = (int)FD->param[1];
         wy = FD->param[2];
         y = UD->y;

         /* make sure ny is odd, otherwise the chief ray is a problem... */
         if (!ny&1)ny++;
         if (wy <= 0.0) return(-1);
         error = GetCellCenter(ny, wy, UD->y, &cy);

         if (error) return(0);

         /* offset the coordinates */
         y -= cy;

         p2 = y*y;
         alpha = 1 - (1+FD->k)*FD->cv*FD->cv*p2;
         if (alpha < 0)
         	{
            /* lens is not big enough to fill the cell, radius is too short */
            /* assume it is plane between lenses */
            alpha = 0.0;
            }
         UD->sag1 = (FD->cv*p2)/(1 + sqrt(alpha));
			/* forget supporting a hyper hemisphere! */
         UD->sag2 = UD->sag1;
      	break;
      case 4:
      	/* ZEMAX wants a paraxial ray trace to this surface */
         /* x, y, z, and the optical path are unaffected, at least for this surface type */
         /* for paraxial ray tracing, the return z coordinate should always be zero. */
         /* paraxial surfaces are always planes with the following normals */

         /* for a lens array, only consider the single lens on axis */
         UD->ln =  0.0;
         UD->mn =  0.0;
         UD->nn = -1.0;
			xpower = 0.0;
         ypower = (FD->n2 - FD->n1)*FD->cv;
         if ((UD->n) != 0.0)
         	{
            (UD->l) = (UD->l)/(UD->n);
            (UD->m) = (UD->m)/(UD->n);

            (UD->l) = (FD->n1*(UD->l) - (UD->x)*xpower)/(FD->n2);
            (UD->m) = (FD->n1*(UD->m) - (UD->y)*ypower)/(FD->n2);

            /* normalize */
            (UD->n) = sqrt(1/(1 + (UD->l)*(UD->l) + (UD->m)*(UD->m) ) );
            /* de-paraxialize */
            (UD->l) = (UD->l)*(UD->n);
            (UD->m) = (UD->m)*(UD->n);
            }
         break;
      case 5:
      	/* ZEMAX wants a real ray trace to this surface */

			/* clear the multiple intercept test flag */
         miss_flag = 0;

         if (FD->cv == 0.0)
         	{
	         outofbounds:;
	         UD->ln =  0.0;
   	      UD->mn =  0.0;
      	   UD->nn = -1.0;
			   if (Refract(FD->n1, FD->n2, &UD->l, &UD->m, &UD->n, UD->ln, UD->mn, UD->nn)) return(-FD->surf);
            return(0);
            }
         /* okay, not a plane. */
         ny = (int)FD->param[1];
         wy = FD->param[2];
         x = UD->x;
         y = UD->y;

         /* make sure ny is odd, otherwise the chief ray is a problem... */
         if (!ny&1)ny++;
         if (wy <= 0.0) return(-1);
         error = GetCellCenter(ny, wy, UD->y, &cy);
         if (error) goto outofbounds;

         try_again:;


         /* offset the coordinates */
         y -= cy;
         z = 0.0;

			a = (UD->n) * (UD->n) * FD->k + (UD->n)*(UD->n) + (UD->m)*(UD->m);
			b = ((UD->n)/FD->cv) - (y) * (UD->m);
			c = (y) * (y);
			rad = b * b - a * c;
			if (rad < 0) return(FD->surf);  /* ray missed this surface */
			if (FD->cv > 0) t = c / (b + sqrt(rad));
			else           t = c / (b - sqrt(rad));
			(x) = (UD->l) * t + (x);
			(y) = (UD->m) * t + (y);
			(z) = (UD->n) * t + (z);
			UD->path = t;

         /* okay, if the ray makes a steep angle of intercept,
         we may actually have hit the wrong element. Check it again! */
         if (miss_flag == 0)
         	{
            double new_cy;

				/* avoid infinite loop */
            miss_flag = 1;

      		/* restore global coordinates prior to test */
   	      UD->y = y + cy;

	         error = GetCellCenter(ny, wy, UD->y, &new_cy);
	         if (error) goto outofbounds;
            if (new_cy != cy)
            	{
               /* we hit the wrong one! */
               /* go back to the tangent plane */
               t = -t;
					(x) = (UD->l) * t + (x);
					(y) = (UD->m) * t + (y);

	   	      y = y + cy;

               cy = new_cy;

               goto try_again;
               }
            }

			zc = (z) * FD->cv;
			rad = zc * FD->k * (zc * (FD->k + 1) - 2) + 1;
			casp = FD->cv / sqrt(rad);
			UD->ln = 0.0;
			UD->mn = (y) * casp;
			UD->nn = -sqrt(1.0 - (UD->mn*UD->mn));

         /* restore coordinates */
         UD->x = x;
         UD->y = y + cy;
         UD->z = z;

         if (Refract(FD->n1, FD->n2, &UD->l, &UD->m, &UD->n, UD->ln, UD->mn, UD->nn)) return(-FD->surf);
         break;
      case 6:
      	/* ZEMAX wants the index, dn/dx, dn/dy, and dn/dz at the given x, y, z. */

         /* This is only required for gradient index surfaces, so return dummy values */
         UD->index = FD->n2;
         UD->dndx = 0.0;
         UD->dndy = 0.0;
         UD->dndz = 0.0;
      	break;
      case 7:
      	/* ZEMAX wants the "safe" data. */
         /* this is used by ZEMAX to set the initial values for all parameters and extra data */
         /* when the user first changes to this surface type. */
         /* this is the only time the DLL should modify the data in the FIXED_DATA FD structure */
         FD->param[1] = 5;
         if (FD->cv == 0.0) FD->param[2] = 100.0;
         else FD->param[2] = 0.5/FD->cv;
         for (i = 3; i <= 8; i++) FD->param[i] = 0.0;
         for (i = 1; i <= 200; i++) FD->xdata[i] = 0.0;
         break;
      case 10:
      	/* scale by factor UD->path */
         FD->param[2] *= UD->path;
         break;
      }
   return 0;
   }

int Refract(double thisn, double nextn, double *l, double *m, double *n, double ln, double mn, double nn)
{
double nr, cosi, cosi2, rad, cosr, gamma;
if (thisn != nextn)
	{
	nr = thisn / nextn;
	cosi = fabs((*l) * ln + (*m) * mn + (*n) * nn);
	cosi2 = cosi * cosi;
	if (cosi2 > 1) cosi2 = 1;
	rad = 1 - ((1 - cosi2) * (nr * nr));
	if (rad < 0) return(-1);
	cosr = sqrt(rad);
	gamma = nr * cosi - cosr;
	(*l) = (nr * (*l)) + (gamma * ln);
	(*m) = (nr * (*m)) + (gamma * mn);
	(*n) = (nr * (*n)) + (gamma * nn);
	}
return 0;
}

int GetCellCenter(int ny, double wy, double y, double *cy)
{
double hwy, ty;
int tny;

*cy = 0.0;

tny = 0;

hwy = 0.5*wy;

/* do cy */
if (fabs(y) > hwy)
	{
   ty = y;
   tny = 0;
   if (y > 0)
   	{
      while(ty >= hwy)
      	{
         tny++;
         ty -= wy;
         }
      }
   else
   	{
      while(ty <= -hwy)
      	{
         tny--;
         ty += wy;
         }
      }
   *cy = tny*wy;
   }

/* are there this many cells? */
if (fabs(tny) > (ny-1)/2) return -1;
else return 0;
}

